.text-updater-node {
  height: 50px;
  border: 1px solid #eee;
  padding: 5px;
  border-radius: 5px;
  background: white;
}

.text-updater-node label {
  display: block;
  color: #777;
  font-size: 12px;
}

.handle-text-left {
  position: relative;
  width: 100px;
  transform: translate(10px, 0%);
  font-size: 6px;
  pointer-events: none;
}

.handle-text-right {
  position: absolute;
  right: 10px;  /* this is the constant value for the right-side spacing, set it as per your requirements */
  width: 30px;  
  font-size: 6px;
  pointer-events: none;
}