
html, body, #root {
	width: 100%;
	height: 100%;
	margin: 0;
	padding: 0;
	box-sizing: border-box;
	font-family: sans-serif;
  }

.react-flow__node:not(.selected) {
    border: 4px solid rgba(0, 0, 0, .25);
}

.react-flow__node.selected {
    border: 4px solid rgba(0, 0, 0, .50);
}

.react-flow__edge-path {
    stroke-width: 2px
}
